#!/usr/bin/env ruby
#####################################################################
####### G15 Amarok Plugin Lyrics v0.1 (released under GPL 2.0) ######
#####################################################################
# Author:                        Robert Riemann                     #
# Report bugs and requests to:   robert.riemann@physik.hu-berlin.de #
# Requires:                      g15composer, amarok with dcop      #
#####################################################################
# The script shows synchronous lyrics for the actual played track   #
# meta infos on screen: title, artist, 3 Lines of lyrics,           #
#####################################################################

##################### Preparations ##################################

$KCODE = 'u' # use utf-8
begin
	require 'getoptlong'
	require 'cgi'
	require 'net/http'
rescue LoadError
	error = 'getoptlong and cgi class for Ruby are required for this script.'
	`kdialog --sorry '#{error}'`
	exit
end

class Array
	def detect_index
		return ( self.length - 1 ) unless yield self.at( self.length - 1 )
		each_with_index {|obj, index| return index if yield(obj) }
	end
end

class String
	def cropLine( maxSize )
		return (self.size > maxSize) ? self[0..maxSize-3] + "..." : self

	end
	def cropLine!( maxSize )
		self.replace "#{self.cropLine maxSize}"
	end
end

class GetoptLong
	def to_hash
		hash = {}
		each { |key, val| hash[key] = val or true }
		hash
	end
end

if ARGV.size > 1 or ARGV[0] =~ /^-/ then
	opts = GetoptLong.new(
		[ "--version",	"-v",	GetoptLong::NO_ARGUMENT ],
		[ "--help",	"-h",		GetoptLong::NO_ARGUMENT ],
		[ "--speed", 	"-s",	GetoptLong::REQUIRED_ARGUMENT ],
		[ "--pipe",		"-p",	GetoptLong::REQUIRED_ARGUMENT ],
		[ "--composer", "-cr",	GetoptLong::REQUIRED_ARGUMENT ]
	).to_hash
else
	opts = {}
end

$speed = opts["--composer"] ? opts["--composer"].to_f : 0.5

################# Get Composer Running ###############################

composer_file = opts["--composer"] ? opts["--composer"] : %x{which g15composer}.chomp
$pipe_file = opts["--pipe"] ? opts["--pipe"] : pipe = "/tmp/g15amarokpluginlyrics"

pid = nil

trap("EXIT") do
	Process.kill( "SIGKILL", pid ) if pid
	Process.kill( "SIGKILL", $pid2 ) if $pid2
end

pid = fork do
	exec "#{composer_file} #{$pipe_file}" unless pid
end
################# Got Composer Running ###############################

################# Prepare Display Methods ###########################

def showMessage( text )
	File.open( $pipe_file , "w" ) do |file|
		buffer = <<EOF # Reads the following lines up to EOF and sends it to our pipe
MC 1
PB 0 0 159 42
TO 1 2 2 1 "#{text}"
DL 1 22 158 22 1
TO 1 24 1 1 "powered by: www.lrcdb.org"
TO 1 34 1 1 "Help by adding your lyrics!"
MC 0
EOF
		file.write buffer
	end
end

$lastinfo = nil
$nextpos = 0
def showLyrics( lines, meta )
	maxSize = 30
	lines.collect! { |line| line.cropLine maxSize }
	sep = (meta["title"] and meta["artist"]) ? " - " : nil
	info = meta["artist"] + sep + meta["title"]
	info += " (" + meta["album"] + ")" if meta["album"]
	$lastinfo = info
	if (info.size > maxSize) then
		pos = (info != $lastinfo) ? 0 : $nextpos
		info = (info + " | " + info)[pos,maxSize]
		$nextpos = (pos + 1) % maxSize
	end

	File.open( $pipe_file , "w" ) do |file|
		buffer = <<EOF # Reads the following lines up to EOF and sends it to our pipe
MC 1
PB 0 0 159 42
TO 1 3 1 1 "#{lines[0]}"
TO 1 13 1 1 "#{lines[1]}"
TO 1 23 1 1 "#{lines[2]}"
DL 1 32 158 32 1
TO 1 34 1 1 "#{info}"
MC 0
EOF
		file.write buffer
	end
end
################### Data handling ###################################
class Lyric
	attr_reader :meta, :size

	private
	def initialize( dataFile )
		a = dataFile.to_a
		@meta = {}
		while a[0] =~ %r{^\[(\w+):(.*)\]} do
			@meta[$1] = $2
			a.shift
		end
		a.shift if a[0].chomp.empty?
		@size = a.length
		@data = []
		@data[0] = [] # array for time in sec
		@data[1] = [] # array for text
		for i in 0...a.length do
			m = %r{^\[(\d+):(\d+).?\d*\](.*)}.match( a[i] )
			@data[0][i] = m[1].to_i*60 + m[2].to_i
			@data[1][i] = m[3].chomp
		end
	end
	
	def getTextById( id )
		return ( id >= 0 and @data[1][id]) ? @data[1][id] : ""
	end
	
	public
	def getTextLines( time )
		p = @data[0].detect_index { |element| element > time }
		puts p.inspect
		i = @data[0].detect_index { |element| element > time } - 1
		
		return [getTextById( i ), getTextById( i+1 ), getTextById( i+2) ]
	end
end

def getDataFromDB( path )
	req = Net::HTTP::Get.new( path )
	res = Net::HTTP.start( 'www.lrcdb.org', 80 ) { |http|
		http.request( req )
	}
	if res.message == "OK" then
		puts res.body
		return res.body
	else
		return nil
	end
end

def fetchData( artist, title )
	showMessage( "Fetching Lyrics...")
	path = "/search.php?artist=#{CGI.escape(artist)}&title=#{CGI.escape(title)}&album=&query=plugin&type=plugin&submit=submit"
	response = getDataFromDB( path )
	if ( response !=  "no match" ) and
	   ( lid = %r{^(?:exact|partial):\s+(\d+)(\s+\w+)?}.match( response )[1] ) and
	   ( dataFile = getDataFromDB( "/lyric.php?lid=#{lid}&astext=yes" ) ) and
	   ( aLyric = Lyric.new( dataFile ) ) and
	   ( aLyric.size > 0 ) then
		aLyric
	else
		return nil
	end
end

####################### Initialize Loops ############################
def renderLyrics
	aLyric = nil
	pid2 = fork do
		unless pid2 then
			loop do
				if aLyric then
					textLines = aLyric.getTextLines( %x{dcop amarok player trackCurrentTime}.to_i )
					showLyrics( textLines, aLyric.meta )
					sleep $speed
				else
		# 			%x{dcop amarok playlist popupMessage "get data"}
					unless not ( $artist.empty? and $title.empty? ) and ( aLyric = fetchData( $artist, $title ) ) then
						showMessage( "No Lyrics Found" )
						sleep
					end
				end
			end
		end
	end
	return pid2
end

def trackChange	
	$artist = %x{dcop amarok player artist}.chomp
	$album = %x{dcop amarok player album}.chomp
	$title = %x{dcop amarok player title}.chomp
	$pid2 = renderLyrics
	
end

def engineStateChange
	status = %x{dcop amarok player status}.to_i
	if status != 1 then
		if $pid2 then
			Process.kill( "SIGKILL", $pid2 ) if $pid2
			Process.wait( $pid2 )
			$pid2 = nil
		end
		if status == 2 then
			trackChange
		else
			showMessage( "not playing" )
		end	
	end
end

# showMessage( "not playing" )
engineStateChange

loop do
	message = gets.chomp #Read message from stdin
	case message
		when /^configure\s*/
			%x{dcop amarok playlist popupMessage "This script does not have configuration options."}

		when /^engineStateChange\s*/
			engineStateChange

    end
end